package Hanoi;

=pod

=head1 NAME

Hanoi - stuff.

=head1 SYNOPSIS

stuff

=head1 DESCRIPTION

stuff

=cut

# Class definition
use strict;
use vars   qw/ $VERSION %FIELDS @EXPORT_OK /;
use fields qw/ _discs _pegs no_of_discs no_of_pegs _fromdisc _todisc /;

# Class dependencies

use Hanoi::Peg qw/ is_peg /;
use Hanoi::Disc qw/ is_disc /;
use Exporter;

use Data::Dumper qw(DumperX);
use Term::ReadLine;

# Inheritence
use base qw/ Exporter /;

$VERSION   = (split(/ /, ' $Version: 1.12 $ '))[2];



########################################################################
sub new {
    my $class = shift;
    $class    = ref($class) || $class || return undef;
    my %args  = @_;
    my $self = [\%FIELDS];

    $self  = bless $self, $class;

    return $self;
}



########################################################################
sub start_game {
    my $self = shift;

    my $no_of_discs = $self->discs();
    unless ($no_of_discs) {
	die "you must set the number of discs\n";
    }

    my $no_of_pegs = $self->pegs();
    unless ($no_of_pegs) {
	die "you must set the number of pegs\n";
    }

    print "Starting game with $no_of_discs discs and $no_of_pegs pegs\n";

    my $i;
    for ($i = $self->pegs(); $i > 0; $i--) {
	print ":$i\n";
	$self->add_peg({size => $i});
    }

    # we add all disks to the first peg
    my $first = @{$self->{_pegs}}[0];

    for ($i = $self->discs(); $i > 0; $i--) {
	print ":$i\n";
	my $added = $self->add_disc({size => $i});
	my $add_to_peg = $first->add_disc($added);
	if (defined $add_to_peg) {
	    print STDERR "added disc to peg\n";
	} else {
	    print STDERR "error adding disc to peg\n";
	}
    }

	  print "\n"; $self->pretty_dump();
    $self->play_game();

#    print STDERR DumperX $self->{_discs};
#    print STDERR DumperX $self->{_pegs};
}



########################################################################
sub play_game {
    my $self = shift;
    my $args = shift;

    while (1) {
      my ($from, $to) = $self->_get_from_to();
      $self->_move({ from => $from, to => $to});
      print "\n"; $self->pretty_dump();
    }
}

sub _get_from_to {
    my $self = shift;

    my @input;

    # yeah yeah, don't have more than 99 pegs, OK?
    my $length = 2;

    @input = ();
    my $ctr = 0;
    while ($ctr < 2) {
	my $prompt = 'Please enter peg to move disc ';
	$prompt .= $ctr ? "to" : "from";
	$prompt .= " (q to quit)\n";
	print $prompt;
	read STDIN, $input[$ctr], $length;
	chomp $input[$ctr];
	last if $input[$ctr] =~ /^q$/i;
	$ctr++;
    }

    return @input;
}



########################################################################
sub add_disc {
    my $self = shift;
    my $args = shift;

    my $disc = Hanoi::Disc->new(Size => $args->{size});

    push @{$self->{_discs}}, $disc;

    return $disc;
}




########################################################################
sub add_peg {
    my $self = shift;
    my $args = shift;

    my $peg = Hanoi::Peg->new();

    push @{$self->{_pegs}}, $peg;

    return $peg;
}



########################################################################
########################################################################
########################################################################
########################################################################
########################################################################



########################################################################
sub discs {
    my $self  = shift;
    my $discs = shift;

    if (defined $discs and (not $self->{no_of_discs})) {
	$self->{no_of_discs} = $discs;
    }

    return $self->{no_of_discs};
}



########################################################################
sub pegs {
    my $self = shift;
    my $pegs = shift;

    if (defined $pegs and (not $self->{no_of_pegs})) {
	$self->{no_of_pegs} = $pegs;
    }

    return $self->{no_of_pegs};
}


########################################################################
sub dump {
    my $self = shift;

    for (my $i = $self->discs()-1; $i>-1; $i--) {
      for (my $j = $self->pegs()-1; $j>-1; $j--) {
          if (defined($self->{_pegs}->[$j]->{discs}->[$i])) {
             print $self->{_pegs}->[$j]->{discs}->[$i]->get_size();
          }
          else { print "-"; }

      }
        print "\n";
    }
}



########################################################################
sub pretty_dump {
    my $self = shift;
    my $biggest = $self->discs();
    my $current;

    for (my $i = $self->discs()-1; $i>-1; $i--) {
      for (my $j = 0; $j < $self->pegs(); $j++) {
          if (defined($self->{_pegs}->[$j]->{discs}->[$i])) {
             $current = $self->{_pegs}->[$j]->{discs}->[$i]->get_size();
             print " "x($biggest-$current)."#"."#"x(($current-1)*2)." "x($biggest-$current);
          }
          else { print " "x($biggest-1)."|"." "x($biggest-1); }

      }
        print "\n";
    }
}




########################################################################
sub _move {
    my $self = shift;
    my $params = shift;

    if (ref($params) ne 'HASH') { die "Need a hash passing in.\n"; }

    my $from = $params->{from} || die "No from!\n";
    my $to   = $params->{to}   || die "No to!\n";

    if ($from > $self->pegs() || $to > $self->pegs()) { die "Invalid Peg(s).\n"; }
    if ($from < 0 || $to < 0) { die "Invalid Peg(s).\n"; }

    if ($self->{_pegs}->[$from-1]->is_empty()) {
        print "You can't move from an empty peg\n";
        return 1;
    }

    # simple case - dest peg is empty and src peg isn't...

    if ($self->{_pegs}->[$to-1]->is_empty()) {
        $self->{_pegs}->[$to-1]->add_disc($self->{_pegs}->[$from-1]->remove_disc());
        return 1;
    }

    # Else compare disc sizes.

    $self->{_fromdisc} = $self->{_pegs}->[$from-1]->remove_disc();
    $self->{_todisc}   = $self->{_pegs}->[$to-1]->remove_disc();

    if ( $self->{_fromdisc}->smaller_than($self->{_todisc}) ) {
         # legal moves. pop pegs...
         $self->{_pegs}->[$to-1]->add_disc($self->{_todisc}); 
         $self->{_pegs}->[$to-1]->add_disc($self->{_fromdisc}); 
         return 1;
    } else {
         # better put the discs back on their pegs anyway
         $self->{_pegs}->[$to-1]->add_disc($self->{_todisc});
         $self->{_pegs}->[$from-1]->add_disc($self->{_fromdisc});
         print "disc is too big for dest. Ha.\n";
    }
}
